/*=============================================================================

	ReShade 3 effect file
    visit facebook.com/MartyMcModding for news/updates

    Simple Bloom 1.0.014
    by Marty McFly / P.Gilcher
    part of qUINT shader library for ReShade 3

    CC BY-NC-ND 3.0 licensed.

=============================================================================*/

/*=============================================================================
	Preprocessor settings
=============================================================================*/

#ifndef SAMPLE_HQ
 #define SAMPLE_HQ 1
#endif

/*=============================================================================
	UI Uniforms
=============================================================================*/

uniform float BLOOM_INTENSITY <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 10.00;
	ui_label = "Bloom Intensity";
	ui_tooltip = "Scales bloom brightness.";
> = 1.0;

uniform float BLOOM_CURVE <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 5.00;
	ui_label = "Bloom Curve";
	ui_tooltip = "Higher values limit bloom to bright light sources only.";
> = 2.75;

uniform float BLOOM_SAT <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 5.00;
	ui_label = "Bloom Saturation";
	ui_tooltip = "Adjusts the color strength of the bloom effect";
> = 2.0;
/*
uniform float BLOOM_DIRT <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 2.00;
	ui_label = "Lens Dirt Amount";
	ui_tooltip = "Applies a dirt mask on top of the original bloom.";
> = 0.0;
*/
uniform float BLOOM_LAYER_MULT_1 <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 1.00;
	ui_label = "Bloom Layer 1 Intensity";
	ui_tooltip = "Intensity of this bloom layer. 1 is sharpest layer, 7 the most blurry.";
> = 0.1;
uniform float BLOOM_LAYER_MULT_2 <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 1.00;
	ui_label = "Bloom Layer 2 Intensity";
	ui_tooltip = "Intensity of this bloom layer. 1 is sharpest layer, 7 the most blurry.";
> = 0.2;
uniform float BLOOM_LAYER_MULT_3 <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 1.00;
	ui_label = "Bloom Layer 3 Intensity";
	ui_tooltip = "Intensity of this bloom layer. 1 is sharpest layer, 7 the most blurry.";
> = 0.3;
uniform float BLOOM_LAYER_MULT_4 <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 1.00;
	ui_label = "Bloom Layer 4 Intensity";
	ui_tooltip = "Intensity of this bloom layer. 1 is sharpest layer, 7 the most blurry.";
> = 0.4;
uniform float BLOOM_LAYER_MULT_5 <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 1.00;
	ui_label = "Bloom Layer 5 Intensity";
	ui_tooltip = "Intensity of this bloom layer. 1 is sharpest layer, 7 the most blurry.";
> = 0.5;
uniform float BLOOM_LAYER_MULT_6 <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 1.00;
	ui_label = "Bloom Layer 6 Intensity";
	ui_tooltip = "Intensity of this bloom layer. 1 is sharpest layer, 7 the most blurry.";
> = 0.6;
uniform float BLOOM_LAYER_MULT_7 <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 1.00;
	ui_label = "Bloom Layer 7 Intensity";
	ui_tooltip = "Intensity of this bloom layer. 1 is sharpest layer, 7 the most blurry.";
> = 0.7;

uniform float BLOOM_TONEMAP_COMPRESSION <
	ui_type = "drag";
	ui_min = 0.00; ui_max = 10.00;
	ui_label = "Bloom Tonemap Compression";
	ui_tooltip = "Lower values compress a larger color range.";
> = 5.0;

/*=============================================================================
	Textures, Samplers, Globals
=============================================================================*/

#include "ReShade.fxh"

texture2D MXBLOOM_BloomTexSource 	{ Width = BUFFER_WIDTH/2; 	Height = BUFFER_HEIGHT/2;    Format = RGBA16F;};
texture2D MXBLOOM_BloomTex1 		{ Width = BUFFER_WIDTH/2; 	Height = BUFFER_HEIGHT/2;    Format = RGBA16F;};
texture2D MXBLOOM_BloomTex2			{ Width = BUFFER_WIDTH/4;  	Height = BUFFER_HEIGHT/4;    Format = RGBA16F;};
texture2D MXBLOOM_BloomTex3 		{ Width = BUFFER_WIDTH/8;  	Height = BUFFER_HEIGHT/8;    Format = RGBA16F;};
texture2D MXBLOOM_BloomTex4 		{ Width = BUFFER_WIDTH/16;  Height = BUFFER_HEIGHT/16;   Format = RGBA16F;};
texture2D MXBLOOM_BloomTex5 		{ Width = BUFFER_WIDTH/32;  Height = BUFFER_HEIGHT/32;   Format = RGBA16F;};
texture2D MXBLOOM_BloomTex6 		{ Width = BUFFER_WIDTH/64;  Height = BUFFER_HEIGHT/64;   Format = RGBA16F;};
texture2D MXBLOOM_BloomTex7 		{ Width = BUFFER_WIDTH/128; Height = BUFFER_HEIGHT/128;  Format = RGBA16F;};

sampler2D sMXBLOOM_BloomTexSource	{ Texture = MXBLOOM_BloomTexSource;	};
sampler2D sMXBLOOM_BloomTex1		{ Texture = MXBLOOM_BloomTex1;		};
sampler2D sMXBLOOM_BloomTex2		{ Texture = MXBLOOM_BloomTex2;		};
sampler2D sMXBLOOM_BloomTex3		{ Texture = MXBLOOM_BloomTex3;		};
sampler2D sMXBLOOM_BloomTex4		{ Texture = MXBLOOM_BloomTex4;		};
sampler2D sMXBLOOM_BloomTex5		{ Texture = MXBLOOM_BloomTex5;		};
sampler2D sMXBLOOM_BloomTex6		{ Texture = MXBLOOM_BloomTex6;		};
sampler2D sMXBLOOM_BloomTex7		{ Texture = MXBLOOM_BloomTex7;		};

/*=============================================================================
	Functions
=============================================================================*/

float3 Downsample(sampler2D bloomTexture, float2 bloomTextureSize, float2 UV)
{
	float3 bloomDownsampled = 0;
	float2 offsetScale = 2.0 / bloomTextureSize;
	float4 kernelUV = 0;

	float2 kernelSmallOffsets = float2(1.0,1.0) * offsetScale;
	float2 kernelLargeOffsets = float2(2.0,2.0) * offsetScale;

	float3 kernelCenter = tex2D(bloomTexture, UV).rgb;

	float3 kernelSmall = 0;

	kernelUV.xy = UV + kernelSmallOffsets;
	kernelSmall += tex2Dlod(bloomTexture, kernelUV).rgb; //++
	kernelUV.x = UV.x - kernelSmallOffsets.x;
	kernelSmall += tex2Dlod(bloomTexture, kernelUV).rgb; //-+
	kernelUV.y = UV.y - kernelSmallOffsets.y;
	kernelSmall += tex2Dlod(bloomTexture, kernelUV).rgb; //--
	kernelUV.x = UV.x + kernelSmallOffsets.x;
	kernelSmall += tex2Dlod(bloomTexture, kernelUV).rgb; //+-

#if SAMPLE_HQ == 0
	return kernelCenter / 5.0	
	      + kernelSmall / 5.0;
#else
	float3 kernelLarge1 = 0;

	kernelUV.xy = UV + kernelLargeOffsets;
	kernelLarge1 += tex2Dlod(bloomTexture, kernelUV).rgb; //++
	kernelUV.x = UV.x - kernelLargeOffsets.x;
	kernelLarge1 += tex2Dlod(bloomTexture, kernelUV).rgb; //-+
	kernelUV.y = UV.y - kernelLargeOffsets.y;
	kernelLarge1 += tex2Dlod(bloomTexture, kernelUV).rgb; //--
	kernelUV.x = UV.x + kernelLargeOffsets.x;
	kernelLarge1 += tex2Dlod(bloomTexture, kernelUV).rgb; //+-

	float3 kernelLarge2 = 0;

	kernelUV.xy = UV;
	kernelUV.x += kernelLargeOffsets.x;
	kernelLarge2 += tex2Dlod(bloomTexture, kernelUV).rgb; //+0
	kernelUV.x -= kernelLargeOffsets.x * 2.0;
	kernelLarge2 += tex2Dlod(bloomTexture, kernelUV).rgb; //-0
	kernelUV.x = UV.x;
	kernelUV.y += kernelLargeOffsets.y;
	kernelLarge2 += tex2Dlod(bloomTexture, kernelUV).rgb; //0+
	kernelUV.y -= kernelLargeOffsets.y * 2.0;
	kernelLarge2 += tex2Dlod(bloomTexture, kernelUV).rgb; //0-

	return kernelCenter * 0.5 / 4.0		
	     + kernelSmall  * 0.5 / 4.0;		
	     + kernelLarge1 * 0.125 / 4.0
	     + kernelLarge2 * 0.25 / 4.0;
#endif
}

float3 Upsample(sampler2D bloomTexture, float2 bloomTexelSize, float2 UV)
{
	float4 kernelUV = 0;

	static const float Upscaleradius = 1.5;
	float4 kernelSmallOffsets;
	kernelSmallOffsets.xy = Upscaleradius * bloomTexelSize;
	kernelSmallOffsets.zw = kernelSmallOffsets.xy * 2;

	float3 kernelCenter = tex2D(bloomTexture, UV).rgb;

	float3 kernelSmall1 = 0;

	kernelUV.xy = UV.xy - kernelSmallOffsets.xy;
	kernelSmall1 += tex2Dlod(bloomTexture, kernelUV).rgb; //--
	kernelUV.x += kernelSmallOffsets.z;
	kernelSmall1 += tex2Dlod(bloomTexture, kernelUV).rgb; //+-
	kernelUV.y += kernelSmallOffsets.w;
	kernelSmall1 += tex2Dlod(bloomTexture, kernelUV).rgb; //++
	kernelUV.x -= kernelSmallOffsets.z;
	kernelSmall1 += tex2Dlod(bloomTexture, kernelUV).rgb; //-+

#if SAMPLE_HQ == 0
	return kernelCenter / 5.0
	     + kernelSmall1 / 5.0;
#else
	float3 kernelSmall2 = 0;

	kernelUV.xy = UV.xy + float2(kernelSmallOffsets.x, 0);
	kernelSmall2 += tex2Dlod(bloomTexture, kernelUV).rgb; //+0
	kernelUV.x -= kernelSmallOffsets.z;
	kernelSmall2 += tex2Dlod(bloomTexture, kernelUV).rgb; //-0
	kernelUV.xy = UV.xy + float2(0, kernelSmallOffsets.y);
	kernelSmall2 += tex2Dlod(bloomTexture, kernelUV).rgb; //0+
	kernelUV.y -= kernelSmallOffsets.w;
	kernelSmall2 += tex2Dlod(bloomTexture, kernelUV).rgb; //0-

	return kernelCenter * 4.0 / 16.0
	     + kernelSmall1 * 1.0 / 16.0
	     + kernelSmall2 * 2.0 / 16.0;
#endif
}

/*=============================================================================
	Pixel Shaders
=============================================================================*/

void PS_BloomPrepass(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	float3 bloom = Downsample(ReShade::BackBuffer, ReShade::ScreenSize, uv);

	float bloom_luma = dot(bloom, 0.333);
	bloom = lerp(bloom_luma, bloom, BLOOM_SAT);
	bloom /= bloom_luma + 1e-3;
	bloom_luma = pow(bloom_luma, BLOOM_CURVE) * BLOOM_INTENSITY * BLOOM_INTENSITY;
	bloom *= bloom_luma;

	color = float4(bloom, 1);
}

void PS_Downsample1(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Downsample(sMXBLOOM_BloomTexSource, ReShade::ScreenSize/2.0, uv),1);
}
void PS_Downsample2(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Downsample(sMXBLOOM_BloomTex1, ReShade::ScreenSize/4.0, uv),1);
}
void PS_Downsample3(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Downsample(sMXBLOOM_BloomTex2, ReShade::ScreenSize/8.0, uv),1);
}
void PS_Downsample4(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Downsample(sMXBLOOM_BloomTex3, ReShade::ScreenSize/16.0, uv),1);
}
void PS_Downsample5(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Downsample(sMXBLOOM_BloomTex4, ReShade::ScreenSize/32.0, uv),1);
}
void PS_Downsample6(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Downsample(sMXBLOOM_BloomTex5, ReShade::ScreenSize/64.0, uv),1);
}
void PS_Downsample7(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Downsample(sMXBLOOM_BloomTex6, ReShade::ScreenSize/128.0, uv),1) * BLOOM_LAYER_MULT_7;
}

void PS_Upsample1(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Upsample(sMXBLOOM_BloomTex7, 128.0 * ReShade::PixelSize, uv), BLOOM_LAYER_MULT_6);
}
void PS_Upsample2(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Upsample(sMXBLOOM_BloomTex6, 64.0 * ReShade::PixelSize, uv), BLOOM_LAYER_MULT_5);
}
void PS_Upsample3(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Upsample(sMXBLOOM_BloomTex5, 32.0 * ReShade::PixelSize, uv), BLOOM_LAYER_MULT_4);
}
void PS_Upsample4(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Upsample(sMXBLOOM_BloomTex4, 16.0 * ReShade::PixelSize, uv), BLOOM_LAYER_MULT_3);
}
void PS_Upsample5(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Upsample(sMXBLOOM_BloomTex3, 8.0 * ReShade::PixelSize, uv), BLOOM_LAYER_MULT_2);
}
void PS_Upsample6(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	color = float4(Upsample(sMXBLOOM_BloomTex2, 4.0 * ReShade::PixelSize, uv), BLOOM_LAYER_MULT_1);
}

void PS_Combine(in float4 pos : SV_Position, in float2 uv : TEXCOORD0, out float4 color : SV_Target0)
{
	float3 bloom = Upsample(sMXBLOOM_BloomTex1, 2.0 * ReShade::PixelSize, uv);
	bloom /= dot(float4(BLOOM_LAYER_MULT_1, BLOOM_LAYER_MULT_2, BLOOM_LAYER_MULT_3, BLOOM_LAYER_MULT_4), 1) + dot(float3(BLOOM_LAYER_MULT_5, BLOOM_LAYER_MULT_6, BLOOM_LAYER_MULT_7), 1);
	color = tex2D(ReShade::BackBuffer, uv.xy);
	
	color.rgb = lerp(color.rgb, bloom.rgb * 128.0, 0.16 * saturate(BLOOM_INTENSITY));
	color.rgb = pow(color.rgb, BLOOM_TONEMAP_COMPRESSION);
	color.rgb = color.rgb / (1.0 + color.rgb);
	color.rgb = pow(color.rgb, 1.0 / BLOOM_TONEMAP_COMPRESSION);
}

/*=============================================================================
	Techniques
=============================================================================*/

technique Bloom
{
        pass
	{
		VertexShader = PostProcessVS;
		PixelShader  = PS_BloomPrepass;
		RenderTarget0 = MXBLOOM_BloomTexSource;
	}

	#define PASS_DOWNSAMPLE(i) pass { VertexShader = PostProcessVS; PixelShader  = PS_Downsample##i; RenderTarget0 = MXBLOOM_BloomTex##i; }

	PASS_DOWNSAMPLE(1)
	PASS_DOWNSAMPLE(2)
	PASS_DOWNSAMPLE(3)
	PASS_DOWNSAMPLE(4)
	PASS_DOWNSAMPLE(5)
	PASS_DOWNSAMPLE(6)
	PASS_DOWNSAMPLE(7)

	#define PASS_UPSAMPLE(i,j) pass {VertexShader = PostProcessVS;PixelShader  = PS_Upsample##i;RenderTarget0 = MXBLOOM_BloomTex##j;ClearRenderTargets = false;BlendEnable = true;BlendOp = ADD;SrcBlend = ONE;DestBlend = SRCALPHA;}

	PASS_UPSAMPLE(1,6)
	PASS_UPSAMPLE(2,5)
	PASS_UPSAMPLE(3,4)
	PASS_UPSAMPLE(4,3)
	PASS_UPSAMPLE(5,2)
	PASS_UPSAMPLE(6,1)

	pass
	{
		VertexShader = PostProcessVS;
		PixelShader  = PS_Combine;
	}
}
